Package: incubate
Title: Parametric Time-to-Event Analysis with Variable Incubation
        Phases
Version: 1.4.1
Date: 2026-06-04
Authors@R: 
    person(given = "Matthias", family = "Kuhn",
           role = c("aut", "cre", "cph"),
           email = "matthias.kuhn@tu-dresden.de",
           comment = c(ORCID = "0000-0003-2868-5155"))
Description: Fit parametric models for time-to-event data that show an initial
    'incubation period', i.e. a variable delay phase where no events occur.
    The delayed Weibull distribution serves as the foundational data model.
    For parameter estimation, different flavours of maximum likelihood
    estimation ('MLE') and the method of maximum product of spacings estimation
    ('MPSE') are implemented. Bootstrap confidence intervals for parameters and
    significance tests in a two group setting are provided.
License: LGPL (>= 3)
Encoding: UTF-8
LazyData: true
Imports: future (>= 1.21), future.apply (>= 1.6), glue (>= 1.4), MASS,
        minqa, purrr (>= 0.3), rlang (>= 0.4), stats, survival, tibble
Suggests: boot, dplyr, future.callr, ggplot2 (>= 3.3), knitr, numDeriv,
        patchwork, rmarkdown, spelling, testthat (>= 3.0.0), tidyr,
        withr
URL: https://gitlab.com/imb-dev/incubate/
BugReports: https://gitlab.com/imb-dev/incubate/-/work_items
Language: en-GB
Config/roxygen2/version: 8.0.0
Config/roxygen2/markdown: TRUE
Config/testthat/edition: 3
Config/Needs/check: spelling
Depends: R (>= 3.6.0)
Collate: 'S3_integration.R' 'aaa.R' 'cpp11.R' 'data.R' 'utils.R'
        'delay_estimation.R' 'delay.R' 'delay_test.R'
        'incubate-package.R' 'zzz.R'
VignetteBuilder: knitr
LinkingTo: cpp11
NeedsCompilation: yes
Packaged: 2026-06-04 16:04:03 UTC; kuhnmat
Author: Matthias Kuhn [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-2868-5155>)
Maintainer: Matthias Kuhn <matthias.kuhn@tu-dresden.de>
Repository: CRAN
Date/Publication: 2026-06-08 18:40:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-06-08 22:09:57 UTC; unix
Archs: incubate.so.dSYM
