# Generated by roxygen2: do not edit by hand

S3method(coef,incubate_fit)
S3method(confint,incubate_fit)
S3method(graphics::lines,incubate_fit)
S3method(logLik,incubate_fit)
S3method(plot,incubate_fit)
S3method(plot,incubate_test)
S3method(print,incubate_fit)
S3method(print,incubate_test)
S3method(simulate,incubate_fit)
S3method(summary,incubate_fit)
S3method(transform,incubate_fit)
S3method(update,incubate_fit)
export(buildDist)
export(delay_model)
export(dexp_delayed)
export(dweib_delayed)
export(mexp_delayed)
export(mweib_delayed)
export(pexp_delayed)
export(power_diff)
export(pweib_delayed)
export(qexp_delayed)
export(qweib_delayed)
export(rexp_delayed)
export(rweib_delayed)
export(test_GOF)
export(test_diff)
importFrom(MASS,boxcox)
importFrom(future,plan)
importFrom(future.apply,future_apply)
importFrom(glue,glue)
importFrom(purrr,chuck)
importFrom(rlang,.data)
importFrom(rlang,`%||%`)
importFrom(stats,coef)
importFrom(stats,simulate)
importFrom(stats,update)
importFrom(survival,Surv)
importFrom(survival,is.Surv)
importFrom(tibble,tibble)
useDynLib(incubate, .registration = TRUE)
